/*
 * GRABCONSOLE - Quick and dirty hack on xconsole to grab /dev/console temporarily
 * Author: Gal Roualland.
 * 
 * Heavily based on xconsole, rights follow:
 * 
 * Copyright (c) 1990  X Consortium
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the X Consortium shall not be
 * used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from the X Consortium.
 * 
 * * Author:  Keith Packard, MIT X Consortium
 */

#define DEFAULT_WAITTIME 5

#include <X11/Xos.h>
#include <sys/stat.h>
#include <signal.h>
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#include <stdio.h>
#undef _POSIX_SOURCE
#else
#include <stdio.h>
#endif
#include <ctype.h>

#ifdef X_NOT_STDC_ENV
extern char *malloc ();

#else
#include <stdlib.h>
#endif

#ifdef MINIX
#define USE_FILE
#define FILE_NAME "/dev/log"
#define read(n,b,s) nbio_read(n,b,s)
#endif

#ifdef ultrix
#define USE_FILE
#define FILE_NAME   "/dev/xcons"
#endif

#ifdef __EMX__
#define USE_FILE
#define FILE_NAME   "/dev/console$"
#define INCL_DOSFILEMGR
#define INCL_DOSDEVIOCTL
#include <os2.h>
#endif

#ifndef USE_FILE
#include    <sys/ioctl.h>
#ifdef hpux
#include    <termios.h>
#endif
#ifdef SVR4
#include    <termios.h>
#include    <sys/stropts.h>	/* for I_PUSH */
#ifdef sun
#include    <sys/strredir.h>
#endif
#endif

#if defined(DGUX)
#include <sys/stropts.h>	/* for I_PUSH */
#include <sys/_int_pty_ioctl.h>
#include <sys/stream.h>
#include <sys/ptem.h>
#endif /* DGUX */

#if defined(TIOCCONS) || defined(SRIOCSREDIR) || defined(Lynx)
#define USE_PTY
static int tty_fd, pty_fd;
static char ttydev[64], ptydev[64];

#endif
#endif

#if !defined(DGUX)		/* No /dev/osm in DG/ux */
#if (defined(SVR4) && !defined(sun)) || (defined(SYSV) && defined(i386))
#define USE_OSM
#include <signal.h>
int osm_pipe ();
static int child_pid;

#endif
#endif /* !DGUX */

static int input = -1;
static int debug = 0;
static int sighupped = 0;
static int sigtermed = 0;

#ifdef USE_PTY
int get_pty (int * pty, int * tty, char * ttydev, char * ptydev);
#endif

#ifdef Lynx
static void 
RestoreConsole ()
{
  int fd;

  if ((fd = open ("/dev/con", O_RDONLY)) >= 0)
    newconsole (fd);
}
#endif

static void
OpenConsole ()
{
  struct stat sbuf;
  
  input = -1;
  
  /* must be owner and have read/write permission */
#if !defined(__NetBSD__) && !defined(__OpenBSD__) && !defined(Lynx) && !defined(__EMX__)
  if (!stat ("/dev/console", &sbuf) &&
      (sbuf.st_uid == getuid ()) &&
      !access ("/dev/console", R_OK | W_OK))
#endif
    {
#ifdef USE_FILE
      input = open (FILE_NAME, O_RDONLY);
      
#ifdef __EMX__
      if (input >= 0) {
	ULONG arg = 1, arglen;
	APIRET rc;
	
	if ((rc = DosDevIOCtl (input, 0x76, 0x4d,
			       &arg, sizeof (arg), &arglen,
			       NULL, 0, NULL)) != 0) {
	  close (input);
	  input = -1;
	}
      }
#endif
#endif
#ifdef USE_PTY
      if (get_pty (&pty_fd, &tty_fd, ttydev, ptydev) == 0) {
#ifdef TIOCCONS
	int on = 1;
	
	if (ioctl (tty_fd, TIOCCONS, (char *) &on) != -1)
	  input = pty_fd;
#else
#ifndef Lynx
	int consfd = open ("/dev/console", O_RDONLY);
	
	if (consfd >= 0) {
	  if (ioctl (consfd, SRIOCSREDIR, tty_fd) != -1)
	    input = pty_fd;
	  close (consfd);
	}
#else
	if (newconsole (tty_fd) < 0)
	  perror ("newconsole");
	else {
	  input = pty_fd;
	  atexit (RestoreConsole);
	}
#endif
#endif
      }
#endif
    }
#ifdef USE_OSM
  /* Don't have to be owner of /dev/console when using /dev/osm. */
  if (input < 0)
    input = osm_pipe ();
#endif
}

static void 
CloseConsole ()
{
  if (input >= 0) {
    close (input);
  }
#ifdef USE_OSM
  if (child_pid > 0) {
    kill (child_pid, SIGTERM);
    child_pid = -1;
  }
#endif
#ifdef USE_PTY
  close (tty_fd);
#endif
}

static void usage (char * prog) {
  printf("Grabconsole 0.1. (by Gal Roualland <gael.roualland@iname.com>)\n"
	 "usage: %s [-f logfile] [-p pidfile] [-x] [-d] [-k] [-t secs] [-h]\n"
	 "\n"
	 "Grabs the system console and hold what's written to it.\n"
	 "Useful when booting in order not to flood the boot terminal.\n"
	 "\n"
	 "Options:\n"
	 " -f logfile    Which file to use to save writes to the system console\n"
	 "               (defaults to /var/log/console, use \"-\" for stdout)\n"
	 " -p pidfile    Where to save PID. (defaults to /var/run/grabconsole.pid.\n"
	 " -n            Do not detach from terminal.\n"
	 " -x            Do not save the console contents, just read from it.\n"
	 " -d            Turns on debug mode and do not detach.\n"
	 " -k            Keeps console log on exit.\n"
	 " -t secs       How many seconds to wait before rewritting to the console.\n"
	 "               (defaults to %d)\n"
	 " -h            Displays this :)\n",
	 prog, DEFAULT_WAITTIME);
  exit(EXIT_SUCCESS);
}

static void grabsig(int sig) {
  if (debug) fprintf(stderr, "received signal %d!\n", sig);
  if (sig==SIGHUP)
    sighupped = 1;
  else if (sig==SIGTERM || sig==SIGINT)
    sigtermed = 1;
}

static void chkarg(char * argv, int arg, int argc, char opt) {
  if (arg+1>=argc) {
    fprintf(stderr,
	    "%s: option requires an argument -- %c\n"
	    "Try `%s -?' for help.\n",
            argv, opt, argv);
    exit(EXIT_FAILURE);
  }
}

int main (int argc, char ** argv) {
  fd_set fds;
  char buffer[4096];
  char pidtxt[20];
  char * logfile = "/var/log/console";
  char * pidfile = "/var/run/grabconsole.pid";
  int pid, pidf;
  
  int keep = 0;
  int trash = 0;
  int goon = 1;
  int waittime = DEFAULT_WAITTIME;
  int detach = 1;
  
  int output = -1;
  
  char * opt;
  int arg = 1;

  while (arg<argc) {
    if (argv[arg][0]=='-' && argv[arg][1]) {
      opt = argv[arg]+1;
      while (*opt) {
        switch (*opt) {
	case 'h':
	  usage (argv[0]);
	case 'f':
	  chkarg(argv[0], arg, argc, *opt);
	  logfile = argv[++arg];
	  break;
	case 'n':
	  detach = 0;
	  break;
	case 't':
	  chkarg(argv[0], arg, argc, *opt);
	  waittime = atoi(argv[++arg]);
	  break;
	case 'p':
	  chkarg(argv[0], arg, argc, *opt);
	  pidfile = argv[++arg];
	  break;
	case 'x':
	  trash = 1;
	  break;
	case 'd':
	  debug = 1;
	  break;
	case 'k':
	  keep = 1;
	  break;
	default:
	  fprintf(stderr,
		  "%s: invalid option -- %c\n"
		  "Try `%s -h' for help.\n",
		  argv[0], *opt, argv[0]);
	  exit(EXIT_FAILURE);
	}
	opt++;
      }
      arg++;
    } else {
      fprintf(stderr, "%s: bad command line\n"
	      "Try `%s -h' for help.\n",
	      argv[0], argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  if (!trash) {
    if (!strcmp(logfile, "-")) {
      if (debug) fprintf(stderr, "outputing to stdout.\n");
      output = dup(STDOUT_FILENO);
    } else {
      if (debug) fprintf(stderr, "opening %s.\n", logfile);
      output = open(logfile, O_TRUNC|O_CREAT|O_WRONLY, 0666);
      if (output<0) {
	perror(logfile);
	exit(EXIT_FAILURE);
      }
    }
  } else
    output = -1;

  if (debug) fprintf(stderr, "opening system console.\n");
  OpenConsole();
  if (input<0) {
    fprintf(stderr, "%s: can't open console.\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  if (debug) detach = 0;
  if (detach) {
    pid = fork();
    if (pid<0) {
      perror("fork");
      exit(EXIT_FAILURE);
    } else if (pid!=0) {
      exit(EXIT_SUCCESS);
    }
    setsid();
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
  }

  pid = getpid();
  if (debug) fprintf(stderr, "saving pid %d to %s.\n", pid, pidfile);
  sprintf(pidtxt, "%d", pid);
  pidf = open(pidfile, O_WRONLY|O_CREAT|O_TRUNC, 0666);
  if (pidf>=0) {
    write(pidf, pidtxt, strlen(pidtxt));
    close(pidf);
  } else
    if (!detach)
      perror(pidfile);
  
  if (debug) fprintf(stderr, "installing signal handler.\n");
  signal(SIGTERM, grabsig);
  signal(SIGHUP, grabsig);
  sighupped = 0;
  sigtermed = 0;
  
  while(goon) {
    int sel;
    
    FD_ZERO(&fds);
    FD_SET(input, &fds);
    sel = select (input + 1, &fds, NULL, NULL, NULL);
    if (sel==1 && FD_ISSET(input, &fds)) {
	int n = read(input, buffer, sizeof(buffer));
	if (n>0 && !trash && (output>=0))
	  write(output, buffer, n);
    }
    if (sighupped) {
      if (debug) fprintf(stderr, "closing console.\n");
      CloseConsole();
      if (debug) fprintf(stderr, "sleeping %d secs.\n", waittime);
      sleep(waittime);
      OpenConsole();
      sighupped = 0;
      if (input<0) {
	if (debug) fprintf(stderr, "console is now opened by other process.\n");
	if (!trash && strcmp(logfile, "-")) {
	  close(output);
	  if (debug) fprintf(stderr, "dumping saved text to system console\n");
	  input = open(logfile, O_RDONLY);
#ifdef USE_FILE
	  output = open(FILE_NAME, O_WRONLY);
#else
	  output = open("/dev/console", O_WRONLY);
#endif
	  if (input>=0 && output>=0) {
	    int n;
	    while ((n = read(input, buffer, sizeof(buffer)))>0)
	      write(output, buffer, n);
	  }
	  close(input);
	  close(output);
	  output = -1;
	  input = -1;
	}
	sigtermed = 1;
      } else
	if (debug) fprintf(stderr, "console still available, resuming normal operations.\n");
    }
    if (sigtermed) {
      CloseConsole();
      if (output>=0)
	close(output);
      if (!keep && strcmp(logfile, "-")) {
	if (debug) fprintf(stderr, "unlinking %s.\n", logfile);
	unlink(logfile);
      }
      unlink(pidfile);
      exit(EXIT_SUCCESS);
    }
  }

  return EXIT_SUCCESS;
}

#ifdef USE_PTY
/* This function opens up a pty master and stuffs it's value into pty.
 * If it finds one, it returns a value of 0.  If it does not find one,
 * it returns a value of !0.  This routine is designed to be re-entrant,
 * so that if a pty master is found and later, we find that the slave
 * has problems, we can re-enter this function and get another one.
 */

/* this is from xterm sources */
#include    "ptyx.h"

#if defined(DGUX)

int get_pty (pty, tty, ttydev, ptydev)
     int *pty, *tty;
     char *ttydev, *ptydev;

{
  int ptyffd;

  if ((*pty = open ("/dev/ptmx", O_RDWR)) < 0) {
    printf ("Couldnt open /dev/ptmx in get_pty(..)\n");
    return 1;
  }
  strcpy (ttydev, (char *) ptsname (*pty));
  setpgrp ();
  grantpt (*pty);
  unlockpt (*pty);
  if ((ptyffd = open (ttydev, O_RDWR)) < 0) {
    printf ("Trying to obtain fildes for ttydev failed!\n");
    if (*pty >= 0) {
      close (*pty);
    }
    return 1;
  }
  if (ioctl (ptyffd, I_PUSH, "ptem") < 0) {
    printf ("Couldnt I_PUSH module ptem\n");
    if (*pty >= 0) {
      close (*pty);
    }
    return 1;
  }
  if (ioctl (ptyffd, I_PUSH, "ldterm") < 0) {
    printf ("Couldnt I_PUSH ldterm \n");
    if (*pty >= 0) {
      close (*pty);
    }
    return 1;
  }
  if (ioctl (ptyffd, I_PUSH, "ttcompat") < 0) {
    printf ("Couldnt I_PUSH ttcompat\n");
  }
  *tty = ptyffd;
  return 0;

}

#else /* Other than DGUX */

int get_pty (pty, tty, ttydev, ptydev)
     int *pty, *tty;
     char *ttydev, *ptydev;
{
#ifdef SVR4
  if ((*pty = open ("/dev/ptmx", O_RDWR)) < 0) {
    return 1;
  }
  grantpt (*pty);
  unlockpt (*pty);
  strcpy (ttydev, (char *) ptsname (*pty));
  if ((*tty = open (ttydev, O_RDWR)) >= 0) {
    (void) ioctl (*tty, I_PUSH, "ttcompat");
    return 0;
  }
  if (*pty >= 0)
    close (*pty);
#else /* !SVR4, need lots of code */
#ifdef USE_GET_PSEUDOTTY
  if ((*pty = getpseudotty (&ttydev, &ptydev)) >= 0 &&
      (*tty = open (ttydev, O_RDWR)) >= 0)
    return 0;
  if (*pty >= 0)
    close (*pty);
#else
  static int devindex, letter = 0;

#if defined(umips) && defined (SYSTYPE_SYSV)
  struct stat fstat_buf;

  *pty = open ("/dev/ptc", O_RDWR);
  if (*pty < 0 || (fstat (*pty, &fstat_buf)) < 0) {
    return (1);
  }
  sprintf (ttydev, "/dev/ttyq%d", minor (fstat_buf.st_rdev));
  sprintf (ptydev, "/dev/ptyq%d", minor (fstat_buf.st_rdev));
  if ((*tty = open (ttydev, O_RDWR)) >= 0) {
    /* got one! */
    return (0);
  }
  close (*pty);
#else /* not (umips && SYSTYPE_SYSV) */
#ifdef CRAY
  for (; devindex < 256; devindex++) {
    sprintf (ttydev, "/dev/ttyp%03d", devindex);
    sprintf (ptydev, "/dev/pty/%03d", devindex);

    if ((*pty = open (ptydev, O_RDWR)) >= 0 &&
	(*tty = open (ttydev, O_RDWR)) >= 0) {
      /* We need to set things up for our next entry * into this * function! */
      (void) devindex++;
      return (0);
    }
    if (*pty >= 0)
      close (*pty);
  }
#else /* !CRAY */
#ifdef sgi
  {
    char *slave;

    slave = _getpty (pty, O_RDWR, 0622, 0);
    if ((*tty = open (slave, O_RDWR)) != -1)
      return 0;
  }
#else
  strcpy (ttydev, "/dev/ttyxx");
  strcpy (ptydev, "/dev/ptyxx");
  while (PTYCHAR1[letter]) {
    ttydev[strlen (ttydev) - 2] = ptydev[strlen (ptydev) - 2] =
      PTYCHAR1[letter];

    while (PTYCHAR2[devindex]) {
      ttydev[strlen (ttydev) - 1] = ptydev[strlen (ptydev) - 1] =
	PTYCHAR2[devindex];
      if ((*pty = open (ptydev, O_RDWR)) >= 0 &&
	  (*tty = open (ttydev, O_RDWR)) >= 0) {
	/* We need to set things up for our next entry * into this  *
	 * function! */
	(void) devindex++;
	return (0);
      }
      if (*pty >= 0)
	close (*pty);
      devindex++;
    }
    devindex = 0;
    (void) letter++;
  }
#endif /* sgi else not sgi */
#endif /* CRAY else not CRAY */
#endif /* umips && SYSTYPE_SYSV */
#endif /* USE_GET_PSEUDOTTY */
#endif /* SVR4 */
  /* We were unable to allocate a pty master!  Return an error * * condition
   * and let our caller terminate cleanly. */
  return (1);
}
#endif
#endif /* DGUX */

#ifdef USE_OSM
/*
 * On SYSV386 there is a special device, /dev/osm, where system messages
 * are sent.  Problem is that we can't perform a select(2) on this device.
 * So this routine creates a streams-pty where one end reads the device and
 * sends the output to xconsole.
 */

#ifdef SCO
#define	OSM_DEVICE	"/dev/error"
#else
#ifdef USL
#define OSM_DEVICE	"/dev/osm2"
#define NO_READAHEAD
#else
#define	OSM_DEVICE	"/dev/osm"
#endif
#endif

#ifdef ISC
#define NO_READAHEAD
#endif

int 
osm_pipe ()
{
  int tty;
  char ttydev[64];

  if (access (OSM_DEVICE, R_OK) < 0)
    return -1;
  if ((tty = open ("/dev/ptmx", O_RDWR)) < 0)
    return -1;

  grantpt (tty);
  unlockpt (tty);
  strcpy (ttydev, (char *) ptsname (tty));

  if ((child_pid = fork ()) == 0) {
    int pty, osm, nbytes, skip;
    char cbuf[128];

    skip = 0;
#ifndef NO_READAHEAD
    osm = open (OSM_DEVICE, O_RDONLY);
    if (osm >= 0) {
      while ((nbytes = read (osm, cbuf, sizeof (cbuf))) > 0)
	skip += nbytes;
      close (osm);
    }
#endif
    pty = open (ttydev, O_RDWR);
    if (pty < 0)
      exit (1);
    osm = open (OSM_DEVICE, O_RDONLY);
    if (osm < 0)
      exit (1);
    for (nbytes = 0; skip > 0 && nbytes >= 0; skip -= nbytes) {
      nbytes = skip;
      if (nbytes > sizeof (cbuf))
	nbytes = sizeof (cbuf);
      nbytes = read (osm, cbuf, nbytes);
    }
    while ((nbytes = read (osm, cbuf, sizeof (cbuf))) >= 0)
      write (pty, cbuf, nbytes);
    exit (0);
  }
  signal (SIGHUP, grabsig);
  signal (SIGINT, grabsig);
  signal (SIGTERM, grabsig);
  return tty;
}
#endif /* USE_OSM */
